import urllib.request
from typing import (
    Any,
    Iterable,
    List,
)

from bs4 import BeautifulSoup

from horsebox.cli import FILENAME_PREFIX
from horsebox.indexer.factory import prepare_doc
from horsebox.indexer.schema import SCHEMA_FIELD_CONTENT
from horsebox.model import TDocument
from horsebox.model.collector import Collector
from horsebox.utils.ipv6 import ipv6_disabled


class CollectorHtml(Collector):
    """
    HTML Collector Class.

    Used to collect the content of an HTML page.
    """

    pages: List[str]

    def __init__(  # noqa: D107
        self,
        pages: List[str],
    ) -> None:
        self.pages = pages

    @staticmethod
    def create_instance(**kwargs: Any) -> Collector:
        """Create an instance of the collector."""
        return CollectorHtml(kwargs['root_path'])

    def collect(self) -> Iterable[TDocument]:
        """
        Collect the data to index.

        Returns:
            Iterable[TDocument]: The collected documents.
        """
        for page in self.pages:
            if page.startswith(FILENAME_PREFIX):
                with open(page[1:], 'r') as file:
                    content = file.read()
            else:
                with ipv6_disabled():
                    with urllib.request.urlopen(page) as response:
                        content = response.read()

            soup = BeautifulSoup(content, 'html.parser')

            name = (
                title[0].get_text()
                if (soup.html and (title := soup.html.find_all('title', limit=1)) and len(title))
                else None
            )

            yield prepare_doc(
                **{
                    'name': name,
                    'path': page,
                    SCHEMA_FIELD_CONTENT: soup.get_text(),
                }
            )



from horsebox.cli.render import (
    Format,
    render_warning,
)
from horsebox.collectors import FILENAME_PIPE
from horsebox.commands import build
from horsebox.indexer.index import open_index
from horsebox.indexer.metadata import get_build_args


def refresh(
    index: str,
    format: Format,
) -> None:
    """
    Refresh an index.

    Args:
        index (str): The location of the persisted index.
        format (Format): The rendering format to use.
    """
    t_index, _ = open_index(
        index,
        format,
        skip_expiration_warning=True,
    )
    if not t_index:
        return

    build_args = get_build_args(index)
    if not build_args:
        render_warning(f'The index {index} has no build arguments')
        return

    build_args.source = list(
        filter(
            lambda s: s != FILENAME_PIPE,
            build_args.source,
        )
    )
    if not build_args.source:
        render_warning(f'The index {index} has no identifiable data source')
        return

    build(
        build_args.source,
        build_args.pattern,
        index,
        build_args.collector_type,
        build_args.collect_as_jsonl,
        format,
    )



from dataclasses import dataclass
from enum import Enum
from typing import List

import tantivy


class FieldType(str, Enum):
    """Types of Fields."""

    TEXT = 'text'
    INTEGER = 'integer'
    DATE = 'date'


@dataclass
class Field:
    """Index Field."""

    name: str
    type: FieldType
    description: str
    stored: bool = True
    indexed: bool = True
    # https://docs.rs/tantivy/latest/tantivy/fastfield/
    fast: bool = False
    # It will chop your text on punctuation and whitespaces,
    # removes tokens that are longer than 40 chars, and lowercase your text.
    # https://docs.rs/tantivy/latest/tantivy/tokenizer/#default
    tokenizer_name: str = 'default'
    """
    'default', 'raw' or 'en_stem'

    See https://docs.rs/tantivy/latest/tantivy/tokenizer/
    """
    index_option: str = 'position'
    """
    'basic', 'freq' or 'position'

    See https://docs.rs/tantivy/latest/tantivy/schema/enum.IndexRecordOption.html
    """


SCHEMA_FIELD_CONTENT = 'content'
SCHEMA_FIELDS: List[Field] = [
    Field('name', FieldType.TEXT, 'Name of the container (file, etc.)'),
    Field('type', FieldType.TEXT, 'Type of the container'),
    # `fast` mode is required to aggregate on `content`
    Field(
        SCHEMA_FIELD_CONTENT,
        FieldType.TEXT,
        'Content of the container',
        # Required to aggregation
        fast=True,
    ),
    Field('path', FieldType.TEXT, 'Full path to the content', indexed=False),
    Field('size', FieldType.INTEGER, 'Size of the content'),
    Field('date', FieldType.DATE, 'Date-time of the content'),
]
DEFAULT_FIELD_NAMES = [
    'name',
    SCHEMA_FIELD_CONTENT,
]
"""
Fields used to search if no field is specifically defined in the query.
By default, all tokenized and indexed fields are default fields.
"""


def get_schema() -> tantivy.Schema:
    """Get the schema of the index."""
    builder = tantivy.SchemaBuilder()

    for field in SCHEMA_FIELDS:
        if field.type == FieldType.TEXT:
            builder = builder.add_text_field(
                field.name,
                stored=field.stored,
                # `fast` added in https://github.com/quickwit-oss/tantivy-py/pull/458
                fast=field.fast,  # type: ignore
                tokenizer_name=field.tokenizer_name,
                index_option=field.index_option,
            )
        elif field.type == FieldType.INTEGER:
            builder = builder.add_integer_field(
                field.name,
                stored=field.stored,
                indexed=field.indexed,
                fast=field.fast,
            )
        elif field.type == FieldType.DATE:
            builder = builder.add_date_field(
                field.name,
                stored=field.stored,
                indexed=field.indexed,
                fast=field.fast,
            )
        else:
            raise NotImplementedError(f'Invalid field type: {field.type}')

    return builder.build()



import re
import unicodedata

__RE_HTML_STRIP = re.compile('<[^>]*>')


def normalize_string(
    text: str,
    keep_accents: bool = False,
) -> str:
    """
    Normalize a string.

    - Replacement of accented characters by their non-accented equivalent.
    - Conversion of Unicode escaped characters.

    Args:
        text (str): The string to normalize.
        keep_accents (bool): Whether the accented characters should be preserved or not.

    Returns:
        str: The normalized string.
            Empty string in case of conversion error (silent fail).
    """
    # Normalize the string
    form = 'NFKC' if keep_accents else 'NFKD'
    normalized = ''.join(
        [
            c
            for c in unicodedata.normalize(form, text)
            if not unicodedata.combining(c)
        ]
    )
    # Convert Unicode escapes
    try:
        unescaped = normalized.encode('utf-16', 'surrogatepass').decode('utf-16')
    except Exception:
        return ''

    return unescaped


def strip_html_tags(text: str) -> str:
    """
    Remove the HTML tags from a string.

    >>> strip_html_tags('<p>lorem ipsum</p>')
    'lorem ipsum'

    >>> strip_html_tags('lorem ipsum')
    'lorem ipsum'

    Args:
        text (str): The string to process.

    Returns:
        str: The cleaned string.
    """
    return __RE_HTML_STRIP.sub('', text).strip()


def strip_spaces(text: str) -> str:
    r"""
    Remove the duplicate spaces from a string.

    >>> strip_spaces('')
    ''

    >>> strip_spaces('lorem  ipsum dolor')
    'lorem ipsum dolor'

    >>> strip_spaces('lorem  ipsum\ndolor\tsit')
    'lorem ipsum dolor sit'

    Args:
        text (str): The string to process.

    Returns:
        str: The cleaned string.
    """
    return ' '.join(text.split())



from itertools import (
    islice,
    chain,
)
from typing import (
    Iterable,
    TypeVar,
)

T = TypeVar('T')


def batched(
    iterable: Iterable[T],
    size: int,
) -> Iterable[Iterable[T]]:
    """
    Batch data from the iterable into tuples of length `size`.

    Adapted from https://docs.python.org/3/library/itertools.html#itertools.batched.

    >>> list(map(lambda b: list(b), batched('ABCDEFG', 3)))
    [['A', 'B', 'C'], ['D', 'E', 'F'], ['G']]

    >>> list(map(lambda b: list(b), batched('ABCDEF', 2)))
    [['A', 'B'], ['C', 'D'], ['E', 'F']]

    >>> list(map(lambda b: list(b), batched('ABC', 1)))
    [['A'], ['B'], ['C']]

    >>> list(map(lambda b: list(b), batched('', 1)))
    []

    Args:
        iterable (Iterable[T]): The iterable to batch.
        size (int): The size of a batch.

    Yields:
        Iterable[Iterable[T]]: The batches.
    """
    if size < 1:
        raise ValueError('size must be at least one')

    iterator = iter(iterable)
    while batch := islice(iterator, size):
        try:
            first = next(batch)
        except StopIteration:
            return
        yield chain((first,), batch)



from enum import Enum
from typing import (
    List,
    Optional,
)

import tantivy

from horsebox.cli.param_parser import (
    parse_params,
    parse_params_group,
)


class TokenizerType(str, Enum):
    """
    Types of Tokenizers.

    See https://docs.rs/tantivy/latest/tantivy/tokenizer/trait.Tokenizer.html.
    """

    RAW = 'raw'
    """For each value of the field, emit a single unprocessed token."""
    SIMPLE = 'simple'
    """Tokenize the text by splitting on whitespaces and punctuation."""
    WHITESPACE = 'whitespace'
    """Tokenize the text by splitting on whitespaces."""
    FACET = 'facet'
    """Process a Facet binary representation and emits a token for all of its parent."""
    REGEX = 'regex'
    """Tokenize the text by using a regex pattern to split.
    
    Args:
        pattern (str)
    """
    NGRAM = 'ngram'
    """
    Tokenize the text by splitting words into n-grams of the given size(s).

    Args:
        min_gram (int) = 2
        max_gram (int) = 3
        prefix_only (bool) = False
    """


class FilterType(str, Enum):
    """
    Types of Filters.

    See https://docs.rs/tantivy/latest/tantivy/tokenizer/trait.TokenFilter.html.
    """

    ALPHANUM_ONLY = 'alphanum_only'
    """Removes all tokens that contain non ascii alphanumeric characters."""
    ASCII_FOLD = 'ascii_fold'
    """
    Converts alphabetic, numeric, and symbolic Unicode characters which are not in the first 127 ASCII characters
    into their ASCII equivalents, if one exists.
    """
    LOWERCASE = 'lowercase'
    """Lowercase terms."""
    REMOVE_LONG = 'remove_long'
    """
    Removes tokens that are longer than a given number of bytes.

    Args:
        length_limit (int)
    """
    STEMMER = 'stemmer'
    """
    Stemmer token filter.

    Tokens are expected to be lowercased beforehand.

    Args:
        language (str)
    """
    STOPWORD = 'stopword'
    """
    Removes stop words for a given language.

    Args:
        language (str)
    """
    CUSTOM_STOPWORD = 'custom_stopword'
    """
    Removes stop words from a given a list.

    Args:
        stopwords (List[str])
    """
    SPLIT_COMPOUND = 'split_compound'
    """
    Splits compound words into their parts based on a given dictionary.

    Args:
        constituent_words: (List[str])
    """


def get_analyzer(
    tokenizer_type: TokenizerType,
    tokenizer_params: Optional[str],
    filter_types: List[FilterType],
    filter_params: Optional[str],
) -> tantivy.TextAnalyzer:
    """
    Create an analyzer.

    Args:
        tokenizer_type (TokenizerType): The type of tokenizer to use.
        tokenizer_params (Optional[str]): The parameters of the tokenizer.
        filter_types (List[FilterType]): The list of filters to apply.
        filter_params (Optional[str]): The parameters of the filters.
    """
    params = parse_params(tokenizer_params, is_raw=tokenizer_type in [TokenizerType.REGEX])

    tokenizer_factory = getattr(tantivy.Tokenizer, tokenizer_type.value)
    t_tokenizer: tantivy.Tokenizer = tokenizer_factory(**params)

    builder = tantivy.TextAnalyzerBuilder(t_tokenizer)

    filter_params_group = parse_params_group(filter_params, len(filter_types))

    for filter_type, filter_param in zip(filter_types, filter_params_group):
        params = parse_params(filter_param)

        filter_factory = getattr(tantivy.Filter, filter_type.value)
        t_filter: tantivy.Filter = filter_factory(**params)
        builder = builder.filter(t_filter)

    return builder.build()



import os
from typing import (
    Any,
    Iterable,
    List,
)

import click
import ijson

from horsebox.cli import (
    FILENAME_PREFIX,
    OPTION_COLLECT_AS_JSONL,
)
from horsebox.indexer.factory import prepare_doc
from horsebox.model import TDocument
from horsebox.model.collector import Collector


class CollectorRaw(Collector):
    """
    Raw Collector Class.

    Used to collect ready to index JSON documents.
    """

    root_path: List[str]
    collect_as_jsonl: bool

    def __init__(  # noqa: D107
        self,
        root_path: List[str],
        collect_as_jsonl: bool,
    ) -> None:
        self.root_path = root_path
        self.collect_as_jsonl = collect_as_jsonl

    @staticmethod
    def create_instance(**kwargs: Any) -> Collector:
        """Create an instance of the collector."""
        return CollectorRaw(
            kwargs['root_path'],
            kwargs[OPTION_COLLECT_AS_JSONL],
        )

    def collect(self) -> Iterable[TDocument]:
        """
        Collect the documents to index.

        Returns:
            Iterable[TDocument]: The collected documents.
        """
        for root_path in self.root_path:
            filename = root_path[1:] if root_path.startswith(FILENAME_PREFIX) else root_path

            with click.open_file(filename, 'r') as file:
                if self.__is_jsonl(filename):
                    # See https://pypi.org/project/ijson/#options-1
                    items = ijson.items(file, '', multiple_values=True)
                else:
                    items = ijson.items(file, 'item')

                for item in items:
                    yield prepare_doc(**item)

    def __is_jsonl(
        self,
        filename: str,
    ) -> bool:
        _, fileext = os.path.splitext(filename)
        return self.collect_as_jsonl or fileext in ['.jsonl', '.ndjson']



from typing import (
    Any,
    Dict,
    List,
    Optional,
    Union,
)

from horsebox.cli.render import render_error

__PARAM_GROUP_SEPARATOR = '|'
__PARAM_SEPARATOR = ';'
__KEY_VALUE_SEPARATOR = '='
__VALUE_STRING_MARKER = "'"
__VALUE_LIST_BEGIN_MARKER = '['
__VALUE_LIST_END_MARKER = ']'
__VALUE_LIST_SEPARATOR = ','


def build_params_group(params: List[str]) -> str:
    """
    Build parameters from a group of parameters.

    >>> build_params_group(['param1=1', 'param2=2'])
    'param1=1|param2=2'

    >>> build_params_group(['', 'param1=1'])
    '|param1=1'

    Args:
        params (List[str]): The group of parameters.
    """
    return __PARAM_GROUP_SEPARATOR.join(params)


def parse_params_group(
    params: Optional[str],
    expected: int,
) -> List[str]:
    """
    Parse groups of parameters from a string.

    >>> parse_params_group('', 3)
    ['', '', '']

    >>> parse_params_group('param1=1|param2=2', 2)
    ['param1=1', 'param2=2']

    >>> parse_params_group('|param1=1', 2)
    ['', 'param1=1']

    Args:
        params (Optional[str]): The string containing the parameters.
        expected (int): The number of expected groups.
    """
    if params:
        params_group = params.split(__PARAM_GROUP_SEPARATOR)
        if len(params_group) != expected:
            render_error(f'Incorrect number of parameters groups: {params} (expected: {expected})')
        else:
            return params_group
    else:
        return [''] * expected

    return []


def parse_params(
    params: Optional[str],
    is_raw: bool = False,
) -> Dict[str, Any]:
    """
    Parse parameters from a string.

    >>> parse_params(None)
    {}

    >>> parse_params('text=lorem;number=123;boolean=true;list=[a,b,c]')
    {'text': 'lorem', 'number': 123, 'boolean': True, 'list': ['a', 'b', 'c']}

    >>> parse_params('list=[a,b,c]', False)
    {'list': ['a', 'b', 'c']}

    >>> parse_params('list=[a,b,c]', True)
    {'list': '[a,b,c]'}

    Args:
        params (Optional[str]): The string containing the parameters.
        is_raw (bool): Whether the value should be returned as is or parsed.
            Default to False.
    """
    if not params:
        return {}

    parsed: Dict[str, Any] = {}

    for param in params.split(__PARAM_SEPARATOR):
        key_val = param.strip().split(__KEY_VALUE_SEPARATOR, 1)
        if len(key_val) != 2:
            render_error(f'Invalid key-value parameter: {param}')

        key, value = key_val
        typed_value = value if is_raw else __parse_typed_value(value)
        if typed_value is None:
            render_error(f"Couldn't detect the type of the value: {param}")

        parsed[key] = typed_value

    return parsed


def __parse_typed_value(value: str) -> Optional[Union[int, str, bool, List[str]]]:
    """
    Parse a string value to a typed one.

    >>> __parse_typed_value('lorem')
    'lorem'

    >>> __parse_typed_value("'lorem'")
    'lorem'

    >>> __parse_typed_value('TRUE')
    True

    >>> __parse_typed_value('true')
    True

    >>> __parse_typed_value('1234')
    1234

    >>> __parse_typed_value('[a,b,c]')
    ['a', 'b', 'c']

    Args:
        value (str): The string value to parse.
    """
    value = value.strip()

    if value.startswith(__VALUE_LIST_BEGIN_MARKER) and value.endswith(__VALUE_LIST_END_MARKER):
        values = value.strip(__VALUE_LIST_BEGIN_MARKER).strip(__VALUE_LIST_END_MARKER).split(__VALUE_LIST_SEPARATOR)
        return list(
            map(
                lambda v: v.strip(__VALUE_STRING_MARKER),
                values,
            )
        )
    elif value.lower() == 'true':
        return True
    elif value.lower() == 'false':
        return False
    elif value.isdigit():
        return int(value)

    return value.strip(__VALUE_STRING_MARKER)



from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    Any,
    Iterable,
)

from horsebox.model import TDocument


class Collector(ABC):
    """Collector Class."""

    @staticmethod
    @abstractmethod
    def create_instance(**kwargs: Any) -> 'Collector':
        """Create an instance of the collector."""
        ...

    @abstractmethod
    def collect(self) -> Iterable[TDocument]:
        """
        Collect the documents to index.

        Returns:
            Iterable[TDocument]: The collected documents.
        """
        ...



from typing import (
    Any,
    List,
    Tuple,
)

import click


class CombinedOption(click.Option):
    """Combined Options Support Class."""

    def __init__(self, *args, **kwargs) -> None:  # noqa: D107
        self.required_if = kwargs.pop('required_if')
        self.ignore_if = kwargs.pop('ignore_if', None)

        super().__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args) -> Tuple[Any, List[str]]:  # noqa: D102
        if (
            # The option can't be ignored due to the presence of another option
            self.ignore_if not in opts
            # The associated option is not provided
            and self.required_if not in opts
        ):
            raise click.UsageError(f'Option {self.required_if} is required with {self.name}')

        return super().handle_parse_result(ctx, opts, args)



import re
from collections import OrderedDict
from time import monotonic_ns
from typing import (
    Any,
    Dict,
    List,
    Optional,
)

import tantivy

from horsebox.cli import (
    ELLIPSIS,
    OPTION_COLLECT_AS_JSONL,
)
from horsebox.cli.config import config
from horsebox.cli.render import (
    Format,
    render,
    render_error,
)
from horsebox.collectors import CollectorType
from horsebox.collectors.factory import get_collector
from horsebox.formatters.text import (
    LINE_BREAK,
    snippet_add_style,
)
from horsebox.indexer.index import (
    feed_index,
    open_index,
)
from horsebox.indexer.metadata import IndexBuildArgs
from horsebox.indexer.schema import (
    DEFAULT_FIELD_NAMES,
    SCHEMA_FIELD_CONTENT,
    get_schema,
)
from horsebox.model import TOutput
from horsebox.utils.stopword import (
    STOP_WORDS_COUNT,
    is_stopword,
)

# Syntax: word~, word~1, word~2
# Ignored: "word1 word2"~ (proximity search)
# Inspired by https://lucene.apache.org/core/2_9_4/queryparsersyntax.html#Fuzzy%20Searches
__RE_IS_FUZZY = re.compile(r"^(?P<word>[^\"']+)~((?P<distance>\d))?$")
__LEVENSHTEIN_DISTANCE_DEFAULT = 1
__LEVENSHTEIN_DISTANCE_MAX = 2

__EMPTY_QUERY = 'Query(EmptyQuery)'


def search(
    source: Optional[List[str]],
    pattern: List[str],
    index: Optional[str],
    collector_type: CollectorType,
    collect_as_jsonl: bool,
    query_string: str,
    limit: int,
    highlight: bool,
    count: bool,
    top: bool,
    format: Format,
) -> None:
    """
    Search an index.

    Args:
        source (Optional[List[str]]): The datasources to search from.
        pattern (List[str]): The pattern to identify the containers to index.
        index (Optional[str]): Location of the index to build/search.
        collector_type (CollectorType): The collector to use for indexing.
        collect_as_jsonl (bool): Whether the JSON documents should be collected as JSON Lines or not.
        query_string (str): The search query.
            See https://docs.rs/tantivy/latest/tantivy/query/struct.QueryParser.html.
        limit (int): the number of search results to return.
        highlight (bool): Whether the keywords found should be highlighted or not.
        count (bool): Whether a count operation should be done or not.
        top (bool): Whether the top list of keywords found should be returned or not.
        format (Format): The rendering format to use.
    """
    # region Open index

    took_index: Optional[int] = None

    if source and pattern:
        # Live search
        collector, extra_args = get_collector(
            collector_type,
            source,
            pattern,
        )

        t_index, took_index = feed_index(
            collector.create_instance(
                root_path=source,
                pattern=pattern,
                **({OPTION_COLLECT_AS_JSONL: collect_as_jsonl} | extra_args),
            ),
            # If an index is provided, the index will be (re) built
            index,
            IndexBuildArgs(
                source=source,
                pattern=pattern,
                collector_type=collector_type,
                collect_as_jsonl=collect_as_jsonl,
            ),
        )
    elif index:
        # Search on an existing index
        t_index, _ = open_index(index, format)
        if not t_index:
            return
    else:
        raise ValueError('`--from` or `--index` must be specified')

    # endregion
    # region Prepare query

    t_query: tantivy.Query = __parse_query_string(
        t_index,
        query_string,
    )

    # endregion
    # region Search

    searcher: tantivy.Searcher = t_index.searcher()

    if top:
        __top_keywords_impl(
            searcher,
            t_query,
            limit,
            format,
            took_index,
        )
    else:
        __search_impl(
            t_index,
            searcher,
            t_query,
            limit,
            highlight,
            count,
            format,
            took_index,
        )

    # endregion


def __parse_query_string(
    t_index: tantivy.Index,
    query_string: str,
) -> tantivy.Query:
    try:
        if match := __RE_IS_FUZZY.match(query_string.strip()):
            # Basic implementation of the fuzzy search through a query string.
            # `FuzzyTermQuery` is not accessible from the query parser:
            # https://github.com/quickwit-oss/tantivy/issues/1112#issuecomment-880331994
            # Attention! Highlight will not work (see https://github.com/quickwit-oss/tantivy/issues/2576)

            distance = min(int(match['distance'] or __LEVENSHTEIN_DISTANCE_DEFAULT), __LEVENSHTEIN_DISTANCE_MAX)
            if (text := match['word']).startswith(f'{SCHEMA_FIELD_CONTENT}:'):
                # The field must not be specified in the query string
                text = text[len(f'{SCHEMA_FIELD_CONTENT}:') :]

            t_query: tantivy.Query = tantivy.Query.fuzzy_term_query(
                get_schema(),
                SCHEMA_FIELD_CONTENT,
                text.strip(),
                distance,
            )
        else:
            # https://docs.rs/tantivy/latest/tantivy/query/struct.QueryParser.html
            # Sample search: https://github.com/quickwit-oss/tantivy-py/blob/master/tests/tantivy_test.py
            t_query: tantivy.Query = t_index.parse_query(
                query_string,
                DEFAULT_FIELD_NAMES,
            )

            if str(t_query) == __EMPTY_QUERY:
                render_error(f'Empty query: {query_string}')
    except ValueError as e:
        render_error(f'Query parse error: {e}')

    return t_query


def __search_impl(
    t_index: tantivy.Index,
    searcher: tantivy.Searcher,
    query: tantivy.Query,
    limit: int,
    highlight: bool,
    count: bool,
    format: Format,
    took_index: Optional[int] = None,
) -> None:
    # region Search

    start = monotonic_ns()
    result: tantivy.SearchResult = searcher.search(query, 1 if count else max(limit, 1))
    took = monotonic_ns() - start

    # endregion
    # region Process result

    if count:
        render(
            {
                'took': {
                    'index': took_index,
                    'search': took,
                },
                'count': result.count,  # type: ignore
            },
            format,
        )
    else:
        snippet_generator: Optional[tantivy.SnippetGenerator] = None
        if highlight:
            # https://docs.rs/tantivy/latest/tantivy/snippet/struct.SnippetGenerator.html
            # https://tantivy-search.github.io/examples/snippet.html
            snippet_generator = tantivy.SnippetGenerator.create(
                searcher,
                query,
                t_index.schema,
                SCHEMA_FIELD_CONTENT,
            )
            snippet_generator.set_max_num_chars(config.highlight_max_chars)

        outputs: TOutput = []
        for score, doc_address in result.hits:
            hit = searcher.doc(doc_address)
            doc = hit.to_dict()

            # region Prepare document for output

            if contents := doc.get(SCHEMA_FIELD_CONTENT):
                if config.render_max_content is None:
                    content = contents[0]
                else:
                    content = contents[0][: config.render_max_content] + ELLIPSIS
            else:
                content = None

            output = OrderedDict(
                name=doc['name'][0],
                content=content,
                path=v[0] if (v := doc.get('path')) else None,
                size=v[0] if (v := doc.get('size')) else None,
                date=v[0] if (v := doc.get('date')) else None,
                score=score,
            )
            if snippet_generator:
                snippet: tantivy.Snippet = snippet_generator.snippet_from_doc(hit)

                highlight_output = (
                    snippet_add_style(snippet) if format == Format.TXT else __build_output_from_snippet(snippet)
                )
                if highlight_output:
                    # Highlights will be generated for search based on text fields.
                    output['highlight'] = highlight_output

            if format == Format.TXT:
                outputs.append(LINE_BREAK)

            # endregion

            outputs.append(output)

        render(
            {
                'took': {
                    'index': took_index,
                    'search': took,
                },
                'hits': outputs,
            },
            format,
        )

    # endregion


def __top_keywords_impl(
    searcher: tantivy.Searcher,
    query: tantivy.Query,
    limit: int,
    format: Format,
    took_index: Optional[int] = None,
) -> None:
    # region Search

    agg_name = 'top_keywords'
    top_count = max(limit, 1)

    start = monotonic_ns()
    result: Dict[str, Any] = searcher.aggregate(
        query,
        {
            agg_name: {
                'terms': {
                    'field': SCHEMA_FIELD_CONTENT,
                    # Extra size to discard potential stop-words.
                    # No extra size for digits, as it would lead to request for too many buckets.
                    'size': top_count + STOP_WORDS_COUNT,
                },
            },
        },
    )
    took = monotonic_ns() - start

    # endregion
    # region Process result

    top_keywords = result[agg_name]

    outputs: TOutput = []
    count = 0
    sum_other_doc_count = 0  # Discarded stop-words count

    for bucket in top_keywords['buckets']:
        key: str = bucket['key']
        if len(key) < config.top_min_chars or key.isdigit() or is_stopword(key):
            # Exclude single characters, digits, stop-words
            sum_other_doc_count += bucket['doc_count']
            continue

        output = OrderedDict(
            key=key,
            doc_count=bucket['doc_count'],
        )
        count += 1

        if format == Format.TXT:
            outputs.append(LINE_BREAK)

        outputs.append(output)
        if count >= limit:
            break

    render(
        {
            'took': {
                'index': took_index,
                'search': took,
            },
            'sum_other_doc_count': top_keywords['sum_other_doc_count'] + sum_other_doc_count,
            'doc_count_error_upper_bound': top_keywords['doc_count_error_upper_bound'],
            'keywords': outputs,
        },
        format,
    )

    # endregion


def __build_output_from_snippet(snippet: tantivy.Snippet) -> Optional[Dict[str, Any]]:
    highlighted = snippet.highlighted()
    if not highlighted:
        return None

    return {
        'html': snippet.to_html(),
        'fragment': snippet.fragment(),
        'ranges': [
            {
                'start': h.start,
                'end': h.end,
            }
            for h in highlighted
        ],
    }



from datetime import datetime
from typing import (
    Any,
    Dict,
    Generator,
    List,
    Optional,
    cast,
)

import tantivy

from horsebox.formatters import DATE_FORMAT
from horsebox.model import TOutput

__SNIPPET_BEGIN = '<b>'
__SNIPPET_END = '</b>'

__COLOR_GREEN = '\033[92m'
__COLOR_RESET = '\033[0m'
__BOLD = '\033[1m'

__INDENT = '  '
__LINE_BREAK = '* * * * *'
LINE_BREAK = {}
"""Line break item."""


def format_txt(output: TOutput) -> Generator[str, Any, None]:
    """
    Text formatter.

    Args:
        output (TOutput): The content to format.

    Yields:
        Generator[str, Any, None]: The formatted content.
    """
    for item in output if isinstance(output, List) else [output]:
        if item == LINE_BREAK:
            yield __LINE_BREAK
            continue

        for field, value in item.items():
            if value is None:
                value = ''
            elif isinstance(value, Dict):
                # Dictionary
                yield f'{field.capitalize()}:'
                for sub_item in format_txt(value):
                    yield f'{__INDENT}{sub_item}'
            elif isinstance(value, List) and len(value) and isinstance(value[0], Dict):
                # List of dictionaries
                for sub_item in format_txt(value):
                    yield sub_item
            else:
                if isinstance(value, float):
                    value = round(value, 3)
                elif isinstance(value, datetime):
                    value = cast(datetime, value).strftime(DATE_FORMAT)
                elif isinstance(value, List):
                    value = ','.join(value)

                yield f'{field.capitalize()}: {value}'


def snippet_add_style(snippet: tantivy.Snippet) -> Optional[str]:
    """
    Add style decorators to a search result snippet.

    Args:
        snippet (tantivy.Snippet): The snippet.

    Returns:
        Optional[str]: The decorated snippet.
    """
    content = snippet.to_html().replace(__SNIPPET_BEGIN, __BOLD + __COLOR_GREEN).replace(__SNIPPET_END, __COLOR_RESET)

    # NOTE The code below has been disabled as the ranges may not be accurate with Unicode (emojis) characters
    """
    highlighted = snippet.highlighted()
    if not highlighted:
        return None

    content = snippet.fragment()
    # Process ranges in reverse order to preserve position information accuracy
    for highlighted in snippet.highlighted()[::-1]:
        content = (
            content[: highlighted.start]
            + __BOLD
            + __COLOR_GREEN
            + content[highlighted.start : highlighted.end]
            + __COLOR_RESET
            + content[highlighted.end :]
        )
    """

    return content



import os
from collections import OrderedDict
from functools import reduce
from typing import Any

import tantivy

from horsebox.cli.render import (
    Format,
    render,
)
from horsebox.indexer.index import open_index
from horsebox.indexer.metadata import get_build_args


def inspect(
    index: str,
    format: Format,
) -> None:
    """
    Inspect an index.

    Args:
        index (str): The location of the persisted index.
        format (Format): The rendering format to use.
    """
    t_index, timestamp = open_index(index, format)
    if not t_index:
        return

    build_args = get_build_args(index)

    searcher: tantivy.Searcher = t_index.searcher()

    size = reduce(
        lambda acc, filename: acc + os.path.getsize(filename),
        os.scandir(index),
        0,
    )

    output: OrderedDict[str, Any] = OrderedDict(
        documents=searcher.num_docs,
        segments=searcher.num_segments,
        size=size,
        timestamp=timestamp,
    )
    if build_args:
        output['using'] = build_args.collector_type
        output['from'] = build_args.source
        output['pattern'] = build_args.pattern
        output['jsonl'] = build_args.collect_as_jsonl
        
    render(output, format)



from typing import (
    List,
    Optional,
    cast,
)

from horsebox.cli import FILENAME_PREFIX
from horsebox.cli.render import (
    Format,
    render,
)
from horsebox.indexer.analyzer import (
    FilterType,
    TokenizerType,
    get_analyzer,
)


def analyze(
    text: str,
    tokenizer_type: TokenizerType,
    tokenizer_params: Optional[str],
    filter_types: List[FilterType],
    filter_params: Optional[str],
    format: Format,
) -> None:
    """
    Analyze a text.

    Args:
        text (str): The text to analyze.
        tokenizer_type (TokenizerType): The tokenizer to use.
        tokenizer_params (Optional[str]): The parameters of the tokenizer.
        filter_types (List[FilterType]): The filters to use.
        filter_params (Optional[str]): The parameters of the filters.
        format (Format): The rendering format to use.
    """
    analyzer = get_analyzer(
        tokenizer_type,
        tokenizer_params,
        filter_types,
        filter_params,
    )

    if text.startswith(FILENAME_PREFIX):
        with open(text[1:], 'r') as file:
            text = file.read()

    analyzed: List[str] = analyzer.analyze(cast(str, text))

    output = {'analyzed': analyzed}

    render(output, format)



import os
from datetime import datetime
from time import monotonic_ns
from typing import (
    Optional,
    Tuple,
)

import tantivy

from horsebox.cli.config import config
from horsebox.cli.render import (
    Format,
    render_error,
    render_warning,
)
from horsebox.indexer.metadata import (
    IndexBuildArgs,
    get_timestamp,
    set_metadata,
)
from horsebox.indexer.schema import get_schema
from horsebox.model.collector import Collector
from horsebox.utils.batch import batched


def feed_index(
    collector: Collector,
    index: Optional[str] = None,
    build_args: Optional[IndexBuildArgs] = None,
) -> Tuple[tantivy.Index, int]:
    """
    Build an index.

    Args:
        collector (Collector): The collector used to collect the documents.
        index (Optional[str]): The path of the index.
            Defaults to None.
        build_args (Optional[IndexBuildArgs]): The arguments used to build the index.
            Defaults to None.

    Returns:
        Tuple[tantivy.Index, int]:
            (the index, the build time).
    """
    documents = collector.collect()

    if index:
        os.makedirs(index, exist_ok=True)

    t_index = tantivy.Index(
        get_schema(),
        index,
        reuse=False,
    )

    num_threads = (os.cpu_count() or 0) // 4
    writer: tantivy.IndexWriter = t_index.writer(num_threads=num_threads)

    start = monotonic_ns()

    for batch in batched(documents, config.index_batch_size):
        for document in batch:
            writer.add_document(tantivy.Document(**document))

        writer.commit()

    writer.wait_merging_threads()

    took = monotonic_ns() - start

    if index:
        set_metadata(
            index,
            datetime.now(),
            build_args,
        )

    # Index must be reloaded for search to work
    t_index.reload()

    return (t_index, took)


def open_index(
    index: str,
    format: Format,
    skip_expiration_warning: bool = False,
) -> Tuple[Optional[tantivy.Index], Optional[datetime]]:
    """
    Open an index.

    Args:
        index (str): The path of the index.
        format (Format): The rendering format to use.
        skip_expiration_warning (bool): Whether the warning on index expiry should be silenced or show.
            Default to False.

    Returns:
        Optional[Tuple[tantivy.Index, Optional[datetime]]]:
            (index object, date of creation of the index).
    """
    exists: bool
    try:
        exists = tantivy.Index.exists(index)
    except ValueError:
        exists = False

    if not exists:
        render_error(f'No index was found at {index}')
        return (None, None)

    t_index = tantivy.Index.open(index)
    timestamp = get_timestamp(index)

    if not skip_expiration_warning and timestamp and format == Format.TXT:
        # Do not render warning in JSON mode, as it may be part of a processing pipeline
        age = datetime.now() - timestamp
        if age > config.index_expiration:
            render_warning(f'Index age limit reached: {str(age).split(".")[0]}')

    return (t_index, timestamp)



from enum import Enum
from typing import (
    Dict,
    NoReturn,
    Optional,
    Union,
)

import click

from horsebox.formatters.json import format_json
from horsebox.formatters.text import format_txt
from horsebox.model import (
    TFormatter,
    TOutput,
)


class Format(str, Enum):
    """Rendering Formats."""

    TXT = 'txt'
    JSON = 'json'


__FORMATTERS: Dict[Format, TFormatter] = {
    Format.TXT: format_txt,
    Format.JSON: format_json,
}


def render(
    output: TOutput,
    format: Format,
) -> None:
    """
    Render an output to the terminal.

    Args:
        output (TOutput): The output to render.
        format (Format): The rendering format to use.
    """
    formatter: Optional[TFormatter] = __FORMATTERS.get(format)
    if not formatter:
        raise ValueError('No formatter found')

    for line in formatter(output):
        click.echo(line)


def render_warning(message: str) -> Union[NoReturn, None]:
    """
    Render a warning message to the terminal.

    Args:
        message (str): The warning message to render.
    """
    click.echo(click.style(message, fg='yellow'))


def render_error(message: str) -> NoReturn:
    """
    Render an error message to the terminal and exit the program.

    Args:
        message (str): The error message to render.
    """
    click.echo(click.style(message, fg='red'), err=True)

    quit(-1)



from typing import (
    Any,
    Dict,
    Optional,
    Tuple,
)

from horsebox.collectors import CollectorType


def guess_html(item: str) -> Optional[Tuple[CollectorType, Dict[str, Any]]]:
    """
    Guess the use of the HTML collector from an item.

    >>> guess_html('file.txt') is None
    True

    >>> guess_html('file.html')
    (<CollectorType.HTML: 'html'>, {})

    Args:
        item (str): The item to check.

    Returns:
        Optional[Tuple[CollectorType, Dict[str, Any]]]:
        - The type of the collector.
        - Some extra arguments to use with the collector.
    """
    if item.endswith('.html'):
        return (CollectorType.HTML, {})

    return None



from collections import OrderedDict
import os
from horsebox.cli.config import CONFIGS
from horsebox.cli.render import (
    Format,
    render,
)
from horsebox.formatters.text import LINE_BREAK
from horsebox.model import TOutput


def config(format: Format) -> None:
    """
    Show the configuration.

    Args:
        format (Format): The rendering format to use.
    """
    outputs: TOutput = []

    for name, config in CONFIGS.items():
        outputs.append(
            OrderedDict(
                name=name,
                description=config.description,
                current=os.environ.get(name, config.default) or '',
                default=config.default or '',
            )
        )
        if format == Format.TXT:
            outputs.append(LINE_BREAK)

    render(outputs, format)



from typing import (
    Any,
    Dict,
)

import click

from horsebox import __version__
from horsebox.cli import (
    OPTION_COLLECT_AS_JSONL,
    PATTERN_ANY,
)
from horsebox.cli.combined_option import CombinedOption
from horsebox.cli.render import Format
from horsebox.collectors import CollectorType
from horsebox.commands import (
    analyze,
    build,
    config,
    inspect,
    refresh,
    schema,
    search,
)
from horsebox.indexer.analyzer import (
    FilterType,
    TokenizerType,
)


def __refine_args(**kwargs: Any) -> Dict[str, Any]:
    if kwargs.pop('json', False):
        kwargs['format'] = Format.JSON

    return kwargs


@click.group(
    context_settings={
        'help_option_names': ['-h', '--help'],
    },
)
@click.version_option(__version__)
def __cli() -> None:
    pass


# region Schema Command


@__cli.command('schema', help='Show the schema of the index.')
@click.option(
    '--json', type=bool, is_flag=True, default=False, help='Format the output as JSON (same as --format json).'
)
@click.option(
    '--format',
    type=click.Choice(list(Format), case_sensitive=False),
    default=Format.TXT,
    show_default=True,
    help='Format of the output.',
)
def __schema_cmd(**kwargs: Any) -> None:
    schema(**__refine_args(**kwargs))


# endregion
# region Configuration Command


@__cli.command('config', help='Show the configuration.')
@click.option(
    '--json', type=bool, is_flag=True, default=False, help='Format the output as JSON (same as --format json).'
)
@click.option(
    '--format',
    type=click.Choice(list(Format), case_sensitive=False),
    default=Format.TXT,
    show_default=True,
    help='Format of the output.',
)
def __config_cmd(**kwargs: Any) -> None:
    config(**__refine_args(**kwargs))


# endregion
# region Build Command


@__cli.command('build', help='Build an index.')
@click.option(
    '--from',
    '-f',
    'source',
    type=click.Path(dir_okay=True, file_okay=True, allow_dash=True),
    multiple=True,
    required=True,
    help='Datasource to index. Prefix filename with @.',
)
@click.option('--pattern', '-p', multiple=True, default=[PATTERN_ANY], show_default=True, help='Containers to index.')
@click.option('--index', '-i', required=True, help='Location where to persist the index.')
@click.option(
    '--using',
    '-u',
    'collector_type',
    type=click.Choice(list(CollectorType), case_sensitive=False),
    default=CollectorType.GUESS,
    show_default=True,
    help='Collector to use for indexing.',
)
@click.option(
    '--jsonl',
    OPTION_COLLECT_AS_JSONL,
    type=bool,
    is_flag=True,
    default=False,
    help='Collect JSON documents as JSON Lines (only with --using raw).',
)
@click.option(
    '--json', type=bool, is_flag=True, default=False, help='Format the output as JSON (same as --format json).'
)
@click.option(
    '--format',
    type=click.Choice(list(Format), case_sensitive=False),
    default=Format.TXT,
    show_default=True,
    help='Format of the output.',
)
def __build_cmd(**kwargs: Any) -> None:
    build(**__refine_args(**kwargs))


# endregion
# region Refresh Command


@__cli.command('refresh', help='Refresh an index.')
@click.option('--index', '-i', required=True, help='Location of the index.')
@click.option(
    '--json', type=bool, is_flag=True, default=False, help='Format the output as JSON (same as --format json).'
)
@click.option(
    '--format',
    type=click.Choice(list(Format), case_sensitive=False),
    default=Format.TXT,
    show_default=True,
    help='Format of the output.',
)
def __refresh_cmd(**kwargs: Any) -> None:
    refresh(**__refine_args(**kwargs))


# endregion
# region Inspect Command


@__cli.command('inspect', help='Inspect an index.')
@click.option('--index', '-i', required=True, help='Location of the index.')
@click.option(
    '--json', type=bool, is_flag=True, default=False, help='Format the output as JSON (same as --format json).'
)
@click.option(
    '--format',
    type=click.Choice(list(Format), case_sensitive=False),
    default=Format.TXT,
    show_default=True,
    help='Format of the output.',
)
def __inspect_cmd(**kwargs: Any) -> None:
    inspect(**__refine_args(**kwargs))


# endregion
# region Search Command


@__cli.command('search', help='Search an index.')
@click.option(
    '--from',
    '-f',
    'source',
    type=click.Path(dir_okay=True, file_okay=True, allow_dash=True),
    multiple=True,
    help='Datasource to index. Prefix filename with @.',
)
@click.option(
    '--pattern',
    '-p',
    type=str,
    multiple=True,
    default=['*'],
    show_default=True,
    help='Containers to index.',
    cls=CombinedOption,
    required_if='source',
    ignore_if='index',
)
@click.option('--index', '-i', help='Location of the index to build/search.')
@click.option(
    '--using',
    '-u',
    'collector_type',
    type=click.Choice(list(CollectorType), case_sensitive=False),
    default=CollectorType.GUESS,
    show_default=True,
    help='Collector to use for indexing.',
)
@click.option(
    '--jsonl',
    OPTION_COLLECT_AS_JSONL,
    type=bool,
    is_flag=True,
    default=False,
    help='Collect JSON documents as JSON Lines (only with --using raw).',
)
@click.option(
    '--query',
    '-q',
    'query_string',
    type=str,
    default='*',
    help='Search query - Syntax: https://docs.rs/tantivy/latest/tantivy/query/struct.QueryParser.html.',
)
@click.option('--limit', '-l', default=10, show_default=True, help='Number of search results.')
@click.option('--highlight', type=bool, is_flag=True, default=False, help='Highlight the found keywords.')
@click.option('--count', type=bool, is_flag=True, default=False, help='Count the number of items found.')
@click.option('--top', type=bool, is_flag=True, default=False, help='Get the top keywords of the result.')
@click.option(
    '--json', type=bool, is_flag=True, default=False, help='Format the output as JSON (same as --format json).'
)
@click.option(
    '--format',
    type=click.Choice(list(Format), case_sensitive=False),
    default=Format.TXT,
    show_default=True,
    help='Format of the output.',
)
def __search_cmd(**kwargs: Any) -> None:
    search(**__refine_args(**kwargs))


# endregion
# region Analyze Command


@__cli.command('analyze', help='Analyze some text.')
@click.option('--text', required=True, help='Text to analyze. Prefix filename with @.')
@click.option(
    '--tokenizer',
    'tokenizer_type',
    type=click.Choice(list(TokenizerType), case_sensitive=False),
    default=TokenizerType.SIMPLE,
    show_default=True,
    help='Tokenizer.',
)
@click.option('--tokenizer-params', 'tokenizer_params', type=str, help='Tokenizer parameters.')
@click.option(
    '--filter',
    'filter_types',
    type=click.Choice(list(FilterType), case_sensitive=False),
    multiple=True,
    default=[],
    show_default=True,
    help='Filter.',
)
@click.option('--filter-params', 'filter_params', type=str, help='Filter parameters.')
@click.option(
    '--json', type=bool, is_flag=True, default=False, help='Format the output as JSON (same as --format json).'
)
@click.option(
    '--format',
    type=click.Choice(list(Format), case_sensitive=False),
    default=Format.TXT,
    show_default=True,
    help='Format of the output.',
)
def __analyze_cmd(**kwargs: Any) -> None:
    analyze(**__refine_args(**kwargs))


# endregion


def main() -> None:  # noqa: D103
    __cli()


if __name__ == '__main__':
    main()



from enum import Enum

FILENAME_PIPE = '-'
"""File name used to represent a data source provided by a stream (pipe)."""

class CollectorType(str, Enum):
    """Type of Collector."""

    FILENAME = 'filename'
    FILECONTENT = 'filecontent'
    FILELINE = 'fileline'
    RSS = 'rss'
    RAW = 'raw'
    HTML = 'html'
    GUESS = 'guess'



import socket
from contextlib import contextmanager
from typing import (
    Any,
    Generator,
    List,
    Tuple,
)


@contextmanager
def ipv6_disabled(*args, **kwds) -> Generator[None, Any, None]:
    """IPv6 disabling context manager."""
    getaddrinfo = socket.getaddrinfo

    def _getaddrinfo_ipv4_only(
        host: str,
        port: int,
        family: int = 0,
        type: int = 0,
        proto: int = 0,
        flags: int = 0,
    ) -> List[Tuple]:
        return getaddrinfo(host, port, socket.AF_INET, type, proto, flags)

    # Disable IPv6
    socket.getaddrinfo = _getaddrinfo_ipv4_only
    try:
        yield
    finally:
        socket.getaddrinfo = getaddrinfo



import html
from datetime import datetime
from time import (
    mktime,
    struct_time,
)
from typing import (
    Any,
    Dict,
    List,
    cast,
)

from feedparser.datetimes import _parse_date

from horsebox.indexer.schema import SCHEMA_FIELD_CONTENT
from horsebox.model import TDocument
from horsebox.utils.normalize import (
    strip_html_tags,
    strip_spaces,
)

__FIELDS_MAP: Dict[str, str] = {
    'name': 'name',
    'type': 'type',
    'content': SCHEMA_FIELD_CONTENT,
    'path': 'path',
    'size': 'size',
    'date': 'date',
}
"""Mapping from parser fields to document ones (argument => schema field)."""


def prepare_doc(**kwargs: Any) -> TDocument:
    """Build a document to be indexed."""
    # If a deterministic `doc_id` is required (for document update/deletion),
    # it could be generated with `abs(hash(kwargs['path']))`

    doc: TDocument = {}

    for arg, field in __FIELDS_MAP.items():
        if value := kwargs.get(arg):
            if field == 'date':
                if isinstance(value, str):
                    # Opportunist use of the datetime parsing function of feedparser
                    dt = _parse_date(value)
                    value = datetime.fromtimestamp(mktime(cast(struct_time, dt))) if dt else None
                elif isinstance(value, struct_time):
                    value = datetime.fromtimestamp(mktime(cast(struct_time, value)))
            elif field == SCHEMA_FIELD_CONTENT:
                value = (
                    [strip_spaces(html.unescape(strip_html_tags(v))) for v in value]
                    if isinstance(value, List)
                    else strip_spaces(strip_html_tags(value))
                )

            if value:
                doc[field] = value

    if not doc.get('size') and (content := doc.get(SCHEMA_FIELD_CONTENT)):
        doc['size'] = len(content)

    return doc



import os
from datetime import datetime
from typing import (
    Any,
    Generator,
    List,
)

from horsebox.collectors.collector_fs.collector import CollectorFS
from horsebox.indexer.factory import prepare_doc
from horsebox.indexer.index import config
from horsebox.model.collector import (
    Collector,
    TDocument,
)
from horsebox.utils.normalize import normalize_string


class CollectorFSByLine(CollectorFS):
    """By Line File System Collector Class."""

    def __init__(  # noqa: D107
        self,
        root_path: List[str],
        pattern: List[str],
    ) -> None:
        super().__init__(
            root_path,
            pattern,
        )

    @staticmethod
    def create_instance(**kwargs: Any) -> Collector:
        """Create an instance of the collector."""
        return CollectorFSByLine(
            kwargs['root_path'],
            kwargs['pattern'],
        )

    def parse(
        self,
        root_path: str,
        file_path: str,
    ) -> Generator[TDocument, Any, None]:
        """
        Parse a file for indexing by its content with line detail.

        Args:
            root_path (str): Base path of the file.
            file_path (str): File to parse.

        Yields:
            Generator[TDocument, Any, None]: The document to index (one document per line).
        """
        stats: os.stat_result = os.stat(file_path)

        filename = file_path[len(root_path) :].strip('/')
        _, ext = os.path.splitext(filename)

        parser_max_line = config.parser_max_line

        with open(
            file_path,
            'r',
            errors='surrogateescape',
        ) as file:
            for pos, line in enumerate(file, start=1):
                if parser_max_line is not None and len(line) > parser_max_line:
                    continue

                line = line.strip()
                if config.string_normalize:
                    line = normalize_string(line)
                if not line:
                    continue

                yield prepare_doc(
                    name=filename,
                    type=ext,
                    content=line,
                    path=f'{file_path}#L{pos}',
                    size=stats.st_size,
                    # Time of most recent content modification expressed in seconds.
                    updated=datetime.fromtimestamp(stats.st_mtime),
                )



from horsebox.collectors.guessers.html import guess_html  # noqa: F401
from horsebox.collectors.guessers.raw import guess_raw  # noqa: F401
from horsebox.collectors.guessers.rss import guess_rss  # noqa: F401



import itertools
import os
from abc import abstractmethod
from glob import iglob
from typing import (
    Any,
    Generator,
    Iterable,
    List,
)

from horsebox.model import TDocument
from horsebox.model.collector import Collector


class CollectorFS(Collector):
    """File System Collector Class."""

    root_path: List[str]
    pattern: List[str]

    def __init__(  # noqa: D107
        self,
        root_path: List[str],
        pattern: List[str],
    ) -> None:
        self.root_path = root_path
        self.pattern = pattern

    def collect(self) -> Iterable[TDocument]:
        """
        Collect the data to index.

        Returns:
            Iterable[TDocument]: The collected documents.
        """
        return itertools.chain.from_iterable(
            # Parse the file...
            self.parse(root_path, filename)
            # ...for each folder...
            for root_path in self.root_path
            # ...for each file in the folder.
            for filename in itertools.chain.from_iterable(
                iglob(os.path.join(os.path.expanduser(root_path), f'**/{p}'), recursive=True) for p in self.pattern
            )
            if os.path.isfile(filename)
        )

    @abstractmethod
    def parse(
        self,
        root_path: str,
        file_path: str,
    ) -> Generator[TDocument, Any, None]:
        """
        Parse a container for indexing.

        Args:
            root_path (str): Base path of the file.
            file_path (str): File to parse.

        Yields:
            Generator[TDocument, Any, None]: The document to index.
        """
        ...



from typing import (
    List,
    Optional,
)

import tantivy

from horsebox.cli.config import config
from horsebox.cli.param_parser import build_params_group
from horsebox.indexer.analyzer import (
    FilterType,
    TokenizerType,
    get_analyzer,
)

__analyzer: Optional[tantivy.TextAnalyzer] = None

_STOP_WORDS_EN_COUNT = 174
"""Number of English stop-words - https://snowballstem.org/algorithms/english/stop.txt"""
_STOP_WORDS_FR_COUNT = 144
"""Number of French stop-words - https://snowballstem.org/algorithms/french/stop.txt"""
STOP_WORDS_COUNT = _STOP_WORDS_EN_COUNT + _STOP_WORDS_FR_COUNT
"""Number of stop-words used in top-keywords analyzer."""


def is_stopword(word: str) -> bool:
    """
    Check if a work is a stop-word or not.

    >>> is_stopword('lorem')
    False

    >>> is_stopword('ipsum')
    False

    >>> is_stopword('and')
    True

    >>> is_stopword('is')
    True

    >>> is_stopword('et')
    True

    >>> is_stopword('il')
    True

    Args:
        word (str): The word to check.
    """
    global __analyzer

    if not __analyzer:
        filter_types: List[FilterType] = [
            FilterType.ALPHANUM_ONLY,
            FilterType.REMOVE_LONG,
            FilterType.LOWERCASE,
            FilterType.STOPWORD,
            FilterType.STOPWORD,
        ]
        filter_params = [
            '',
            'length_limit=40',
            '',
            'language=english',
            'language=french',
        ]

        if config.custom_stopwords:
            filter_types.append(FilterType.CUSTOM_STOPWORD)
            filter_params.append(f'stopwords=[{config.custom_stopwords}]')

        __analyzer = get_analyzer(
            TokenizerType.RAW,
            None,
            filter_types,
            build_params_group(filter_params),
        )

    analyzed: List[str] = __analyzer.analyze(word)

    return len(analyzed) == 0  # The list will be empty for a stop-word



from typing import List

from horsebox.cli import OPTION_COLLECT_AS_JSONL
from horsebox.cli.render import Format
from horsebox.collectors import CollectorType
from horsebox.collectors.factory import get_collector
from horsebox.commands.inspect import inspect
from horsebox.indexer.index import feed_index
from horsebox.indexer.metadata import IndexBuildArgs


def build(
    source: List[str],
    pattern: List[str],
    index: str,
    collector_type: CollectorType,
    collect_as_jsonl: bool,
    format: Format,
) -> None:
    """
    Build a persistent index.

    Args:
        source (List[str]): Locations from which to start indexing.
        pattern (List[str]): The containers to index.
        index (str): The location where to persist the index.
        collector_type (CollectorType): The collector to use.
        collect_as_jsonl (bool): Whether the JSON documents should be collected as JSON Lines or not.
        format (Format): The rendering format to use.
    """
    collector, extra_args = get_collector(
        collector_type,
        source,
        pattern,
    )

    feed_index(
        collector.create_instance(
            root_path=source,
            pattern=pattern,
            **({OPTION_COLLECT_AS_JSONL: collect_as_jsonl} | extra_args),
        ),
        index,
        IndexBuildArgs(
            source=source,
            pattern=pattern,
            collector_type=collector_type,
            collect_as_jsonl=collect_as_jsonl,
        ),
    )

    inspect(index, format)



from horsebox.collectors.collector_fs.fs_by_content import CollectorFSByContent  # noqa: F401
from horsebox.collectors.collector_fs.fs_by_filename import CollectorFSByFilename  # noqa: F401
from horsebox.collectors.collector_fs.fs_by_line import CollectorFSByLine  # noqa: F401



from horsebox.commands.analyze import analyze  # noqa: F401
from horsebox.commands.build import build  # noqa: F401
from horsebox.commands.config import config  # noqa: F401
from horsebox.commands.inspect import inspect  # noqa: F401
from horsebox.commands.refresh import refresh  # noqa: F401
from horsebox.commands.schema import schema  # noqa: F401
from horsebox.commands.search import search  # noqa: F401
